#ifndef IUPOP3_GENERAL_H
#define IUPOP3_GENERAL_H
/*
 ===========================================================================
 =                                                                         =
 =  (C) Copyright 1991-1994 The Trustees of Indiana University             =
 =                                                                         =
 =  Permission to use, copy, modify, and distribute this program for       =
 =  non-commercial use and without fee is hereby granted, provided that    =
 =  this copyright and permission notice appear on all copies and          =
 =  supporting documentation, the name of Indiana University not be used   =
 =  in advertising or publicity pertaining to distribution of the program  =
 =  without specific prior permission, and notice be given in supporting   =
 =  documentation that copying and distribution is by permission of        =
 =  Indiana University.                                                    =
 =                                                                         =
 =  Indiana University makes no representations about the suitability of   =
 =  this software for any purpose. It is provided "as is" without express  =
 =  or implied warranty.                                                   =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 = File:                                                                   =
 =   IUPOP3_GENERAL.H                                                      =
 =                                                                         =
 = Authors:                                                                =
 =   Jacob Levanon & Larry Hughes                                          =
 =   Indiana University                                                    =
 =   University Computing Services, Network Applications                   =
 =                                                                         =
 = Credits:                                                                =
 =   This software is based on the Post Office Protocol version 3,         =
 =   as implemented by the University of California at Berkeley.           =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 =  This file based on version 1.8-1 of the iupop3 server, the             =
 =  last version made made Indiana University.                             =
 =                                                                         =
 =  For differences to this version read changes.txt                       =
 =                                                                         =
 = M. Stenns         Jan 1997                                              =
 =                                                                         =
 ===========================================================================
*/

typedef struct dsc$descriptor_s STR_DESCR;

#define MAX_THREADS 30  /* VMS callable mail limit minus one */

#define TRUE            1
#define FALSE           0
#define NEGATIVE	-1
#define ENDMULTLINTRANS ".\r\n"

#define MAXHOSTNAMELEN  255
#define MAXUSERNAMELEN  12
#define MAXLINELEN      1024
#define MAXMSGLINELEN   1024
#define MAXARGCOUNT     10
#define MAXARGLEN       10

#define TIMESTAMP_FORMAT "<V%s@%s %s>"

#if defined  MAIL_FOLDER || defined USE_MAIL_FOLDER
#define MAX_MESSAGES    5000
#ifndef USE_MAIL_FOLDER
#define USE_MAIL_FOLDER
#endif
#ifdef MAIL_FOLDER
#undef MAIL_FOLDER
#endif
#else /* !USE_MAIL_FOLDER */
#define MAX_MESSAGES    20
#endif /* MAIL_FOLDER */

#define  NEWMAIL_FOLDER  "NEWMAIL"
#define  MAIL_FOLDER     "MAIL"

#define DECNET_MESSAGE 1
#define SMTP_MESSAGE 2

#define MAX_AUTHENTICATION_ATTEMPTS 2

#define APOP_FILENAME "POP_SECRET.DAT"
#define MIN_PASSWORD_LENGTH 8

#define ATTN_SLEEP      0
#define ATTN_CONNECT    1
#define ATTN_DISCONNECT 2
#define ATTN_TIMEOUT    3
#define ATTN_DATA       4
#define ATTN_UNBLOCK    5

#ifdef KERBEROS
#define RFC_PORT        1109
#else
#define RFC_PORT        110
#endif

#define POP_OK          "+OK"
#define POP_ERR         "-ERR"
#define POP_SUCCESS     1
#define POP_FAILURE     0
#define POP_TERMINATE   '.'
#define LOG_ERROR  0
#define LOG_INFO   1
#define LOG_THREAD 2
#define LOG_DEBUG  3

#define ITEM_LENGTH 255

#define READ_BUFFERSIZE 256
/* SEND_BUFFERSIZE defines max. supported length of smtp headers */
#ifdef NETLIB
#define SEND_BUFFERSIZE 7680  
#else
#define SEND_BUFFERSIZE 8192  
#endif

#define SOCK_BUFSIZE 8192

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define pop_command         pop_args[0]	     /*  1st arg is a POP command   */
#define pop_subcommand      pop_args[1]      /*  if 1st token was XTND, then
					         the 2nd is a subcommand    */

typedef enum {					/*  POP processing states */
    auth1,					/*  Authorization	  */
    auth2,					/*  Authorization	  */
    trans,					/*  Transaction	          */
    update,					/*  Update		  */
    halt,					/*  Halt		  */
    error					/*  Error		  */
} state;

typedef struct {                                /*  POP command state info    */
    state       ValidCurrentState;              /*  Commans's operating state */
    char	*command;                       /*  The POP command	      */
    int         min_args;                       /*  Minimum args for command  */
    int         max_args;                       /*  Maximum args for command  */
    int         (*function) ();                 /*  Function for command      */
    state       PostState[2];                   /*  Post-command state	      */
} state_table;

typedef struct {                                /*  Table of extensions       */
    char	*subcommand;                    /*  The POP XTND subcommand   */
    int         min_args;                       /*  Minimum args to xtnd comm.*/
    int         max_args;                       /*  Maximum args to xtnd comm.*/
    int         (*function) ();                 /*  Function for command      */
    char        *validate;                      /*  Validation string         */
} xtnd_table;

typedef struct {                                /*  Message information       */
    int         number;                         /*  Message number	      */
    long        length;                         /*  Length of message (bytes) */
    int         lines;                          /*  Number of lines/message   */
    unsigned int bindate[2];                    /*  message arrival date as 
                                                    a quadword binary value   */
    int         del_flag;                       /*  Is marked for deletion?   */
    int         retr_flag;                      /*  Is marked as retrieved    */
    int         retr_ack;			/*  Is on when retr was 
                                                    acknowledged	      */
    int         foreign_msg;
} Message;

struct IOSB
{
  unsigned short status;
  unsigned short terminator_offset;
  unsigned short terminator;
  unsigned short terminator_size;
};

typedef struct 
{                                               /*  POP parameter block	      */
    char                user[MAXUSERNAMELEN+1]; /*  Name of the POP user      */
    char		*pop_args[MAXARGCOUNT]; /*  Parse POP parameter list  */
    char		*client;                /*  Client's name	      */
    char		ipaddr[16];             /*  Client's IP address	      */
    unsigned short      ipport;                 /*  Client's port	      */
    struct IOSB         read_iosb;              /*  read I/O status block     */
    struct IOSB         write_iosb;             /*  write I/O status block    */
    char                read_buffer[READ_BUFFERSIZE];
    char                command_buffer[READ_BUFFERSIZE+1];
    char                *command_pointer;
    char                previous_command[32];
    unsigned short      msg_count;              /*  Number of new messages    */
    state               CurrentState;           /*  The current POP state     */
    Message		*mptr;                  /*  Message information list  */
    int                 msgs_deleted;           /*  Number mesgs for deletion */
    int                 last_msg;               /*  Last message processed    */
    long                bytes_deleted;          /*  Number of bytes to delete */
    long                newmail_size;           /*  Size of NEWMAIL in bytes  */
    int                 arg_count;              /*  Number of args to parse   */
    int                 sockfd;                 /*  socket in use             */
    int                 decnet;                 /*  FALSE, if from internet,
                                                    ignored, if 
                                                    IGNORE_MAIL11_HEADERS 
						    not set                   */
    char 		*folder_name;		/*  current folder name       */
    unsigned short	channel;		/*  VMS Mail specific */
    unsigned int        user_context;           /*  VMS Mail specific */
    unsigned int        file_context;           /*  VMS Mail specific */
    unsigned int        message_context;        /*  VMS Mail specific */
    int                 newmail_selected;       /*  VMS Mail specific */
    char                mail_directory[256];    /*  VMS Mail specific */
    int                 in_use;                 /*  Multithread specific */
    int                 threadnum;              /*  Multithread specific */
    int                 connected;              /*  Multithread specific */
    int                 attn_state;             /*  state of current thread   */
    unsigned int        start_time[2];          /*  start time (binary)       */
    unsigned int        timeout[2];             /*  client timeout (binary)   */
    int                 authentication_attempts;
    int                 has_no_mail_file;
    struct
    {
      char command[MAXARGLEN+1];
      int  completed;
      int  blocked;
      int  message_id;
      int  message_lines;
      int  lines_sent;
      char *send_buffer;
      char buffer[ITEM_LENGTH+4];
      int  bufsize;
      int  send_bufsize;
      int  send_octets;
      int  tries;
      FILE * ext_file;
      int Start[2];
    } retrieve;
#ifdef NETLIB
    struct SINDEF sin;
#else
    struct sockaddr_in sin;
#endif
#ifdef KERBEROS
    AUTH_DAT            kdata;
    Key_schedule        schedule;
    KTEXT_ST            ticket;
    char                instance[INST_SZ];
    char                version[9];
#endif /* KERBEROS */
} POP;

#ifdef KERBEROS
char local_realm[REALM_SZ];
#endif /* KERBEROS */

/* 
** Prototypes for functions in iupop3_commands.c
*/
int         pop_dele (POP *p);
int         pop_last (POP *p);
int         pop_list (POP *p);
int         pop_uidl (POP *p);
int         pop_parse (POP *p, char *buf);
int         pop_pass (POP *p);
int         pop_quit (POP *p);
int         pop_rset (POP *p);
int         pop_send (POP *p);
int         pop_stat (POP *p);
int         pop_noop (POP *p);
int         pop_updt (POP *p);
int         pop_user (POP *p);
int         pop_apop (POP *p);
int         pop_xtnd (POP *p);
int         pop_xtnd_client (POP *p);
int         pop_xtnd_shutdown (POP *p);
int         pop_xtnd_internals (POP *p);
int         pop_xtnd_stats (POP *p);
int         pop_xtnd_loglevel (POP *p);
int	    pop_xtnd_options (POP *p);
int	    pop_xtnd_validate (char *user, char *command);
int         display_runtime_options (POP *p);
int         pop_authcheck (POP *p, int status);
xtnd_table  *get_pop_subcommand (POP *p);
state_table *pop_get_command(POP *p, char *mp);

/* 
** Prototypes for functions in iupop3_utility.c
*/
void	get_date(char *date_str);
void	get_time(char *time_str);
void	make_argv(char *string, int *argc, char *argv[], char *delimiters);
int	pop_build_info (POP *p);
void	patch_from_line (POP *p, char *from_line);
void	pop_log (int log_level, POP * p, const char *format, ...);
int	pop_msg(POP *p, int stat, const char *format, ...);
void	system_log (int log_level, char *format, ...);
int	netprintf (POP *p, char *format, ...);
int     netwrite_send_buffer (POP *p);
int	sndbuf_cpy (POP *p);
int	strncasecmp (char *str1, char *str2, int len);
int	allocate_send_buffer (POP *p);
char	*lower (char *str);
char	*upper (char *str);
int	check_digest (unsigned int start_time[2], char *secret, char *digest_client);
void    process_enviroment_options (void);
STR_DESCR *Descr (char *String);
/* routines only for netlib: */
int	async_write_netlib  (POP *p, void * astadr, int astparam);
int	write_netlib (int socket, char *message, int buflen);
STR_DESCR *Fixed_Descr (char *Buffer, size_t buflen);


/* 
** Prototypes for functions in iupop3_vms.c
*/
char	*date_vms2unix(char *vmsdate);
float	get_cpu (void);
int	mail_close_file_context (POP *p);
int	mail_close_message_context (POP *p);
int	mail_close_user_context (POP *p);
int	mail_delete_message (POP *p,int *message_id);
int	mail_folder_select (POP *p,char *folder_name);
int	mail_message_info (POP *p,int *message_id);
int	mail_message_bytes (POP *p,int *message_id,unsigned int *bytes);
int	mail_message_lines(POP *p,int *message_id, unsigned int * bindate);
int	mail_open_file_context(POP *p);
int	mail_open_message_context(POP *p);
int	mail_open_user_context(POP *p);
int	mail_retrieve_message(POP *p);
int	mail_retrieve_message_headers (POP *p);
int	mail_retrieve_message_text(POP *p);
int	mail_user_info(POP *p);
int	valid_vms_user(POP *p);
char	*vms_message(unsigned int code);
int	mail_message_move(POP *p,int message_id);
int	mail_message_new_count(POP *p);
int	mail_purge_waste(POP *p);
int	get_shared_secret (POP *p, char * shared_secret, int len);
char    *time_from_bintime (unsigned int bintime[2], int mode);


/* 
** Prototypes for functions in iupop3.c
*/
int	cancel_client_timer (int threadnum);
int	close_pop_thread (POP *p);
void	client_read_ast (int threadnum);
void	client_timeout_ast (int threadnum);
void	client_unblock_ast (int threadnum);
int	create_initial_socket (int pop_port);
int	init_pop_thread (int threadnum, POP *p, int sockfd);
int	client_authenticated (POP *p);
int	netwrite_async (POP *p, char * message, int buflen);
int	issue_client_read_qio (int threadnum);
int	issue_client_timeout_timer (int threadnum);
int	issue_client_unblock_timer (int threadnum);
int	issue_new_connect_qio (void);
void	new_connect_ast (void);
int	open_log_file (int modus);
int	issue_logfile_flush_timer (void);
void	logfile_flush_ast (void);
int     process_new_connect (int threadnum);
void	process_thread (POP *p);
#endif /* IUPOP3_GENERAL_H */
